% Implementation of (Generalized) Isotonic Recursive Partitioning as described in:
% Generalized Isotonic Regression by Luss and Rosset, arXiv:1104.1779
% 
% Compute:
% minimize_y  sum_i(x_i - y_i)^2 
% subject to 
%             x(i) <= x(j) for all (i,j) such that y(i)<=y(j)
% where y is an input vector 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Inputs:  
% A:        n x n matrix s.t. A(i,j)==1 if y(i)<=y(j) and A(i,j)==0 otherwise
% y:        n x 1 input vector to isotonic regression, e.g. observations
% options: options.est_type          'L2' for L2 isotonic regression
%                                    'poisson' for poisson isotonic regression
%                                    'pnorm' for Lp isotonic regression with p>1
%                                    'huberloss' for huber isotonic regression
%                                    'L2_weighted' for L2 weighted isotonic regression
%                                    'huberloss_weighted' for weighted huber isotonic regression
%                                    'poisson_weighted' for weighted poisson regression
%          options.lossFunction      'mse' for mean squared error loss on predictions
%                                    'poisson' for negative poisson loglikelihood on predictions
%          options.delta              delta parameter for (weighted) huber isotonic regression
%          options.weights            vector of input weights for all weighted isotonic regressions
%          options.maxiters:          maximum number of iterations
%          options.check_isotonicity  1 if to check isotonicity, 0 otherwise
%          options.maxNumPartitions   maximum number of partitions to return in the GIRP path
%          options.licenseLP          1 if mosek license, 2 if to use Matlab graphmaxflow in Biostatistics toolbox (default=1)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Outputs:  
% yfits:             global solution to isotonic regression
% cutInds:           groups of indices representing the path of groups that
%                        can be used to recover a path of isotonic solutions
% iter:              number of iterations, i.e. partitions
% violation_flag:    0 if no violation of isotonicity along the path, 1 if there is a violation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Modified by Ronny Luss, May 2011
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [yfits,cutInds,iter,violation_flag] = GIRP(y,A,options)
if ~isfield(options,'est_type'), options.est_type='L2'; end
if ~isfield(options,'lossFunction'), options.lossFunction='mse'; end
if ~isfield(options,'delta'), options.delta=1; end
if ~isfield(options,'weights'), options.weights=ones(length(y),1); end
if ~isfield(options,'maxiters'), options.maxiters=inf; end
if ~isfield(options,'check_isotoncity'), options.check_isotonicity=0; end
if ~isfield(options,'maxNumPartitions'), options.maxNumPartitions=inf; end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Algorithm sketch:
% 1. Start with all nodes in one group
% 2. For each group, compute the estimator of the group (minimizer with no isotonic constraints)
% 3. For each group, compute a partition by solving the linear program (partitioning problem)
% 4. If a group is optimal, fix it.
% 5. Make the partition that has the lowest objective value of the partitioning problem
% 6. Return to Step 2.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

n=length(y);           % number of observations
group=zeros(n,1);      % vector of group IDs.  each group will have a different ID.
cutvalues=zeros(n,1);  % variable for deciding which cut to make at each iterations, i.e. objective value of partitioning problems
cutgroups1=zeros(n,1); % for each possible cut, cutgroups1 has 2 IDs, one for below cut, one for above cut
cutgroups2=zeros(n,1); % for each possible cut, cutgroups2 has same ID for union of observations below and above cut
cutcount1=0;
cutcount2=0;
iter=1;
x=GIRP_cutproblem_wrapper(y,A,options,1:n);  % find an initial partitioning by solving the partitioning problem
yfits=zeros(n,1);      % initialize fits
cutInds=[];            % output variable to track of cuts along the path to the final solution
violation_flag=0;      % output flag to check that isotonicity is maintained

% we check the first possible cut
if sum(x>.9)<n && sum(x<-.9)<n  % if all x(i)~=1 for all i or x(i)~=-1 for all i then split the groups
    % indleft are indices of observations below the cut
    % indright are indices of observations above the cut
    indleft=find(x<=-.9);indright=find(x>=.9); 
    estimator=GIRP_estimator(y([indleft;indright]),options,[indleft;indright]); % compute the estimator of group with all observations
    cutvalues([indleft;indright])=sum(cut_value(y(indright),estimator,options,indright))-sum(cut_value(y(indleft),estimator,options,indleft)); % keep track of objective value to partitioning problem
    cutgroups1(indleft)=cutcount1+1; % cutcount1 now has two different IDs for observations below/above the cut
    cutgroups1(indright)=cutcount1+2;
    cutgroups2([indleft;indright])=cutcount2+1; % keep a unique ID for the union of the observations
    cutcount1=cutcount1+2; % next ID for separated groups is +2 ahead
    cutcount2=cutcount2+1; % next ID for unions is +1 ahead
    maxCutValue=0; % initialize maximum cut value
else
    maxCutValue=-inf; % if maximum cut value is not greater than -infinity, we will know to stop
end

count=1; % counter for new group IDs
countModels=1; % count the number of models/partitions along the isotonic path
cutInds{countModels,1}=[1:n]';  % cutInds{i,1} are indices below cut for the ith group being partitioned 
cutInds{countModels,2}=[];      % cutInds{i,1} are indices above cut for the ith group being partitioned
yfits=GIRP_estimator(y,options,1:n)*ones(n,1); % compute fits of first group (all oberservations in one group) 

countModels=countModels+1; % now start with the second model
while maxCutValue > -inf && iter<options.maxiters-1 && countModels<=options.maxNumPartitions
    
    [maxCutValue,indCutValue]=max(cutvalues);  % cut the group with highest cutvalue into two and then make two new additional possible cuts 
    indsToCut=find(cutgroups2==cutgroups2(indCutValue)); % indices of all nodes in the group to be cut
    temp=sort(unique(cutgroups1(indsToCut)),'ascend');
    if maxCutValue > -inf    % then not all groups are optimal yet so make a new cut
        for i=1:length(temp) % for each of the two new groups from the cut we are now making
            indsToCut=find(cutgroups1==temp(i));
            groupsize=length(indsToCut);
            group(indsToCut)=count; % divide these groups in the current solution
            count=count+1;   % this counter leaves all optimal groups with distinct values

            cutInds{countModels,i}=indsToCut;     % choose indices of either observations below or above the cut
            yfits(indsToCut)=GIRP_estimator(y(indsToCut),options,indsToCut); % compute the estimator of this new group to be
        
            % now we make a "future" possible cut (to be added to the possible new groups made latter)
            % NOTE: this is also used to check if the new group is optimal!
            [indToCut_sorted,ind]=sort(indsToCut,'ascend');
            x=GIRP_cutproblem_wrapper(y(indToCut_sorted),A(indToCut_sorted,indToCut_sorted),options,indToCut_sorted);
            x=x(ind); % put the indices back in the original order
 
            % this IF makes a new possible cut if the group was not optimal (see comments above for first initial cut)
            if sum(x>.9)<groupsize && sum(x<-.9)<groupsize % if all x(i)~=1 for all i or x(i)~=-1 for all i then split the groups
                indleft=find(x<=-.9);indright=find(x>=.9);
                estimator=GIRP_estimator(y(indsToCut([indleft;indright])),options,indsToCut([indleft;indright]));
                cutvalues(indsToCut([indleft;indright]))=sum(cut_value(y(indsToCut(indright)),estimator,options,indsToCut(indright)))-sum(cut_value(y(indsToCut(indleft)),estimator,options,indsToCut(indleft)));
                cutgroups1(indsToCut(indleft))=cutcount1+1;
                cutgroups1(indsToCut(indright))=cutcount1+2;
                cutgroups2([indsToCut(indleft);indsToCut(indright)])=cutcount2+1;
                cutcount1=cutcount1+2;
                cutcount2=cutcount2+1;

            else % group is optimal so cutvalues are set accordingly and fits are thus fixed
                cutvalues(indsToCut)=-inf;
            end
        end
        countModels=countModels+1; % move on to the next partition
    end
    
    if options.check_isotonicity
        ret=check_isotonicity(yfits,A); % check isotonicity of the intermediate solutions
        if ret==0
            disp(['Error: Break in isotonocity at iteration ',num2str(iter)]);
            violation_flag=1;
        end
    else
        violation_flag=0;
    end
    iter=iter+1;
end