% Wrapper function to GIRP partitioning (i.e. cut) problem
% For efficiency, the dual problem (the max flow problem) is solved.
% NOTE: the current implementation requires a Mosek license
% Compute:
% minimize_y  sum_i(x_i*gradF(y_i)) 
% subject to 
%             x(i) <= x(j) for all (i,j) such that y(i)<=y(j)
%             0 <= x <= 1
% where y is an input vector and gradF(y_i) is the derivative of the loss
% function with respect to y_i and evaluated at the group estimator
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Inputs:  
% A:        n x n matrix s.t. A(i,j)==1 if y(i)<=y(j) and A(i,j)==0 otherwise
% y:        n x 1 input vector to isotonic regression, e.g. observations
% options:  same options as to GIRP(.) function
% indsALL:  indices of observations in the group to be cut (for use when there are weights)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Outputs:  
% xret:         solution to partitioning problem
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Modified by Ronny Luss, May 2011
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [x_ret] = GIRP_cutproblem_wrapper(y,A,options,indsALL)


[I,J]=find(A);  % find the indices for isotonicity constraints
m=length(I);    % number of isotonicity constraints
n=length(y);    % number of observations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% the dual problem is:
% maximize      sum_i(s_i+t_i)
% subject to 
%               sum_{j:(i,j) in I}(lambda_ij)-sum_{j:(j,i) in I}(lambda_ji)-s_i+t_i=c_i
%               s,t,lambda>=0
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
estimator=GIRP_estimator(y,options,indsALL);
c=cut_value(y,estimator,options,indsALL); % parameter to network flow constraints

x_ret=zeros(n,1);
Aineq=sparse(I,[1:m]',ones(m,1),n,m)-sparse(J,[1:m]',ones(m,1),n,m); % coefficients for network flow constraints
Aineq=[speye(n),-1*speye(n),Aineq];  % appended to coefficients for s and t  

if max(abs(c))>0, c=c/max(abs(c)); end  % normalized for numerical reasons
x=GIRP_LP_subproblem_network_dual_mosek(Aineq,c,m,n);
clear('Aineq');
clear('IR_LP_subproblem_network_dual_mosek');

indleft=find(x<=-.9);indright=find(x>=.9);
x_ret(indleft)=-1; % create return structure (below cut)
x_ret(indright)=1; % create return structure (above cut)
